/*
    Copyright (C) 2007    Dmitry Ivanov, vonami@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <locale.h>

#include "partitionmanager.h"
#include "imager.h"

using namespace std;
using namespace PartimageNG;

void usage()
{
        cerr <<
                "Usage:  partimage-ng [options] command [arguments]\n" <<
                "\n" <<
                "OPTIONS:\n" <<
                "    -b, --blocks number       Read/write the 'number' of adjacent filesystem\n" <<
                "                              data blocks at once. This improves performance.\n" <<
                "    -m, --exclude-mbr         Don't save the MBR boot code\n" <<
                "    -g, --exclude-grub        Don't save the GRUB boot code\n" <<
                "    -s, --full-swap           Save the entire swap partition, not just the\n" <<
                "                              filesystem structures.\n" <<
                "    -t, --description text    Add 'text' as description to the image file\n" <<
                "                              (max 65536 bytes)\n" <<
                "    -f, --force               Force the operation even if the device is\n" <<
                "                              mounted or busy\n" <<
                "    -i, --ignore-bad-sectors  Skip bad sectors when creaing an image of\n" <<
                "                              the device\n" <<
                "    -h, --help                Print this usage message and exit\n" <<
                "\n" <<
                "COMMANDS:\n" <<
                "\n" <<
                "    save device file          Create an image of the 'device' and save it\n" <<
                "                              to 'file'. If 'file' is specified as 'stdout'\n" <<
                "                              the image will be piped to standart output\n" <<
                "                              channel.\n" <<
                "    restore file device       Restore the image from 'file' to 'device'. If\n" <<
                "                              'file' is specified as 'stdin' the image will\n" <<
                "                              be read from standart input channel.\n" <<
                "    info device               Print information about the ''device' like\n" <<
                "                              device geometry, data block size, total number\n" <<
                "                              of data blocks, number of used blocks and\n" <<
                "                              so on. Useful for debugging.\n" <<
                "    imageinfo file            Print information about the image stored in\n" << 
                "                              'file' like type of the image file,\n" <<
                "                              description, number of partitions stored in\n" <<
                "                              the image file, detailed info about every\n" <<
                "                              stored partition and so on. Useful when you\n" <<
                "                              don't remember where this image file came from.\n" <<
                "    dumpbitmap device file    Dump the partition bitmap to a 'file'. Useful\n" <<
                "                              for debugging and reporting bugs\n" <<
                "    dumrawbitmap device file  Dump the internal representation of\n" <<
                "                              partition's bitmap to a 'file'. Useful for\n" <<
                "                              debugging and reporting bugs.\n" <<
                "\n" <<
                "EXAMPLES:\n" <<
                "\n" <<
                "    partimage-ng -b 32 save /dev/sdb ~/sdb.img\n" <<
                "        This will create an image of the whole disk \'/dev/sdb\'\n" <<
                "        (including partition table) and save it to the file \'~/sdb.img\'\n" <<
                "        Also partimage-ng will try to read 32 adjacent filesystem blocks\n" <<
                "        at once to speed things up\n" <<
                "\n" <<
                "    partimage-ng save /dev/sdb1 stdout | gzip -c > ~/sdb1.img.gz\n" <<
                "        This will create an image file of the device \'/dev/sdb1\'\n" <<
                "        and pipe it into gzip to compress\n" <<
                "\n" <<
                "    zcat ~/sdb1.img.gz | partimage-ng restore stdin /dev/sdb1\n" <<
                "        This will restore filesystem on the device \'/dev/sdb1\'\n" <<
                "        from the image \'~/sdb1.img.gz\' piped through gzip to decompress\n";
}

int main(int argc, char **argv)
{
        char *op = 0;
        char *ifile = 0, *ofile = 0;
        char *tail;
        unsigned long desiredBuffering= 0;
        bool saveMbr = true;
        bool grubHack = true;
        bool saveFullSwap = false;
        char *desc = 0;
        bool force = false;
        IOErrorsHandling ioErrorsHandling = StopIOErrors;

        static const char *sopt = "sgmb:t:fih";
        static const struct option lopt[] = {
                { "full-swap",          no_argument,            0, 's' },
                { "exclude-grub",       no_argument,            0, 'g' },
                { "exclude-mbr",        no_argument,            0, 'm' },
                { "blocks",             required_argument,      0, 'b' },
                { "description",        required_argument,      0, 't' },
                { "force",              no_argument,            0, 'f' },
                { "ignore-bad-sectors", no_argument,            0, 'i' },
                { "help",               no_argument,            0, 'h' },
                { 0, 0, 0, 0 }
        };

        // set locale properly
        setlocale(LC_ALL, "");

        int c;
        while ( (c = getopt_long(argc, argv, sopt, lopt, 0)) != -1 ) {
                switch (c) {
                        case 'b':
                                errno = 0;
                                desiredBuffering = strtoul(optarg, &tail, 0);
                                if ( *tail != '\0' ) {
                                        cerr << "Option -b requires an integer argument" << endl;
                                        return 1;
                                }
                                else if ( errno || desiredBuffering > 65535 ) {
                                        cerr << "Number of buffers should not be greater than " << 65535 << endl;
                                        return 1;
                                }
                                else if ( desiredBuffering == 0 ) {
                                        cerr << "Number of buffers should be greater than or equal 1" << endl;
                                        return 1;
                                }
                                break;
                        case 'm':
                                saveMbr = false;
                                break;
                        case 'g':
                                grubHack = false;
                                break;
                        case 's':
                                saveFullSwap = true;
                                break;
                        case 't':
                                desc = optarg;
                                break;
                        case 'f':
                                force = true;
                                break;
                        case 'i':
                                ioErrorsHandling = IgnoreIOErrors;
                                break;
                        case 'h':
                                usage();
                                return 0;
                        case '?':
                                usage();
                                return 1;
                       default:
                                cerr << "Unknown option \'" << argv[optind-1] << "\'" << endl;
                                usage();
                                return 1;
                }
        }

        //  parse arguments
        if ( optind < argc ) {
                op = argv[optind++];
        }
        else {
                usage();
                return 1;
        }

        if ( strcmp(op, "save") == 0 ) {
                if ( optind + 2 != argc ) {
                        cerr << "'save' operation requires 2 arguments" << endl;
                        usage();
                        return 1;
                }
                ifile = argv[optind++];
                ofile = argv[optind++];
        }
        else if ( strcmp(op, "restore") == 0 ) {
                if ( optind + 2 != argc ) {
                        cerr << "'restore' operation requires 2 arguments" << endl;
                        usage();
                        return 1;
                }
                ifile = argv[optind++];
                ofile = argv[optind++];
        }
        else if ( strcmp(op, "info") == 0 ) {
                if ( optind + 1 != argc ) {
                        cerr << "'info' operation requires only 1 argument" << endl;
                        usage();
                        return 1;
                }
                ifile = argv[optind++];
        }
        else if ( strcmp(op, "imageinfo") == 0 ) {
                if ( optind + 1 != argc ) {
                        cerr << "'imageinfo' operation requires only 1 argument" << endl;
                        usage();
                        return 1;
                }
                ifile = argv[optind++];
        }
        else if ( strcmp(op, "dumpbitmap") == 0 ) {
                if ( optind + 2 != argc ) {
                        cerr << "'dumpbitmap' operation requires 2 arguments" << endl;
                        usage();
                        return 1;
                }
                ifile = argv[optind++];
                ofile = argv[optind++];
        }
        else if ( strcmp(op, "dumprawbitmap") == 0 ) {
                if ( optind + 2 != argc ) {
                        cerr << "'dumprawbitmap' operation requires 2 arguments" << endl;
                        usage();
                        return 1;
                }
                ifile = argv[optind++];
                ofile = argv[optind++];
        }
        else {
                cerr << "Unknown operation \'" << op << "\'" << endl;
                return 1;
        }


        //  lets go do it!
        Imager imager;
        imager.setSaveMbr(saveMbr);
        imager.setGrubHack(grubHack);
        imager.setSaveFullSwap(saveFullSwap);
        imager.setIOErrorsHandling(ioErrorsHandling);

        if ( desiredBuffering > 1 ) {
                imager.setDesiredBuffering((u_int16_t)desiredBuffering);
        }

        if ( desc != 0 ) {
                imager.setDescription(desc);
        }

        if ( strcmp(op, "save") == 0 ) {
                if ( PartitionManager::isDrive(ifile) ) {
                        if ( !force && PartitionManager::isBusy(ifile) ) {
                                cerr << "The specified drive is mounted or busy in some other way" << endl;
                                cerr << "Please use the -f (--force) option to ignore that" << endl;
                                return 1;
                        }
                        cerr << "Saving the whole drive " << ifile << endl;
                        return imager.saveDisk(ifile, ofile);
                }
                else if ( PartitionManager::isPartition(ifile) ) {
                        if ( !force && PartitionManager::isBusy(ifile) ) {
                                cerr << "The specified partition is mounted or busy in some other way" << endl;
                                cerr << "Please use the -f (--force) option to ignore that" << endl;
                                return 1;
                        }
                        cerr << "Saving the partition " << ifile << endl;
                        return imager.savePartition(ifile, ofile);
                }
                else {
                        cerr << "Specified input file hasn't been recognized as a valid block device" << endl;
                        cerr << "Check if you have permissions to access this device" << endl;
                        return 1;
                }
        }
        else if ( strcmp(op, "restore") == 0 ) {
                if ( PartitionManager::isDrive(ofile) ) {
                        if ( !force && PartitionManager::isBusy(ofile) ) {
                                cerr << "The specified drive is mounted or busy in some other way" << endl;
                                cerr << "Please use the -f (--force) option to ignore that" << endl;
                                return 1;
                        }
                        cerr << "Restoring the whole drive " << ofile << endl;
                        return imager.restoreDisk(ifile, ofile);
                }
                else if ( PartitionManager::isPartition(ofile) ) {
                        if ( !force && PartitionManager::isBusy(ofile) ) {
                                cerr << "The specified partition is mounted or busy in some other way" << endl;
                                cerr << "Please use the -f (--force) option to ignore that" << endl;
                                return 1;
                        }
                        cerr << "Restoring the partition " << ofile << endl;
                        return imager.restorePartition(ifile, ofile);
                }
                else {
                        cerr << "Specified input file hasn't been recognized as a valid block device" << endl;
                        cerr << "Check if you have permissions to access this device" << endl;
                        return 1;
                }
        }
        else if ( strcmp(op, "info") == 0 ) {
                if ( !PartitionManager::isPartition(ifile) ) {
                        cerr << "Specified input file hasn't been recognized as a valid partition" << endl;
                        cerr << "Check if you have permissions to access this device" << endl;
                        return 1;
                }

                PartitionManager::instantiatePartition(ifile);
                return 0;
        }
        else if ( strcmp(op, "imageinfo") == 0 ) {
                imager.imageinfo(ifile);
                return 0;
        }
        else if ( strcmp(op, "dumpbitmap") == 0 ) {
                cerr << "Dumping the bitmap of " << ifile << " to " << ofile << endl;

                if ( !PartitionManager::isPartition(ifile) ) {
                        cerr << "Specified input file hasn't been recognized as a valid partition" << endl;
                        cerr << "Check if you have permissions to access this device" << endl;
                        return 1;
                }

                if ( !force && PartitionManager::isBusy(ifile) ) {
                        cerr << "The specified partition is mounted or busy in some other way" << endl;
                        cerr << "Please use the -f (--force) option to ignore that" << endl;
                        return 1;
                }

                Partition *part = PartitionManager::instantiatePartition(ifile);
                if ( part == 0 ) {
                        cerr << "Couldn't open partition " << ifile << endl;
                        return 1;
                }

                int dumpd = open(ofile, O_CREAT | O_EXCL | O_WRONLY, 0644);
                if ( dumpd == -1 ) {
                        perror("Couldn't create a dump file");
                        delete part;
                        return 1;
                }

                if ( writen(dumpd, part->bitmap()->bitmap, part->bitmap()->allocsize) == -1 ) {
                        perror("Writing the bitmap failed");
                        return 1;
                }

                delete part;
                close(dumpd);
        }
        else if ( strcmp(op, "dumprawbitmap") == 0 ) {
                cerr << "Dumping the internal bitmap of " << ifile << " to " << ofile << endl;

                if ( !PartitionManager::isPartition(ifile) ) {
                        cerr << "Specified input file hasn't been recognized as a valid partition" << endl;
                        cerr << "Check if you have permissions to access this device" << endl;
                        return 1;
                }

                if ( !force && PartitionManager::isBusy(ifile) ) {
                        cerr << "The specified partition is mounted or busy in some other way" << endl;
                        cerr << "Please use the -f (--force) option to ignore that" << endl;
                        return 1;
                }

                const Partition *part = PartitionManager::instantiatePartition(ifile);
                if ( part == 0 ) {
                        cerr << "Couldn't open partition " << ifile << endl;
                        return 1;
                }

                int dumpd = open(ofile, O_CREAT | O_EXCL | O_WRONLY, 0644);
                if ( dumpd == -1 ) {
                        perror("Couldn't create a dump file");
                        delete part;
                        return 1;
                }

                part->dumpInternalBitmap(dumpd);

                delete part;
                close(dumpd);
        }
        else {
                cerr << "Unknown operation \'" << op << "\'" << endl;
                return 1;
         }

        return 0;
}
