/*
    Copyright (C) 2007    Dmitry Ivanov, vonami@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cstdlib>
#include <iostream>
#include <errno.h>

#include "bitshash.h"
#include "ntfspartition.h"

extern "C" {
        #include "ntfspartition_c.h"
}

using namespace std;
using namespace PartimageNG;

NtfsPartition::NtfsPartition(const PedPartition *pedPartition)
        : Partition(pedPartition), m_usedBlocks(0)
{
        if ( (m_ntfs_vol = partimage_ng_ntfs_open(path())) == 0 ) {
                cerr << "Couldn't open the filesystem" << endl;
                return;
        }

        m_bitmap.bitmap = partimage_ng_ntfs_bitmap(m_ntfs_vol);
        m_bitmap.allocsize = partimage_ng_ntfs_bitmap_allocated_size(m_ntfs_vol);
        if ( m_bitmap.bitmap == 0 ) {
                cerr << "Couldn't read the bitmap" << endl;
                return;
        }

        m_blockSize = partimage_ng_ntfs_block_size(m_ntfs_vol);
        if ( m_blockSize == 0 ) {
                cerr << "Couldn't determine the block size" << endl;
                return;
        }

        m_blocksCount = partimage_ng_ntfs_blocks_count(m_ntfs_vol);
        if ( m_blocksCount == 0 ) {
                cerr << "Couldn't determine the blocks count" << endl;
                return;
        }

        //  + 1 because ntfs stores a copy of boot block at the end of the partition
        //  though, in most cases it will be an incomplete block (always 512 byte long)
        if ( m_blocksCount & 0x07 ) {
                // we have spare bits in the last byte of the bitmap
                m_blocksCount += 1;
                m_bitmap.bitmap[m_blocksCount >> 3] |= 1 << ((m_blocksCount) & 0x07);
        }
        else if ( (m_blocksCount >> 3) + 1 <= m_bitmap.allocsize ) {
                // we have spare bytes in the bitmap, use them
                m_blocksCount += 1;
                m_bitmap.bitmap[m_blocksCount >> 3] = 0xff;
        }
        else {
                // we reached end of the bitmap, realloc it
                m_blocksCount += 1;
                m_bitmap.bitmap = (u_int8_t *) realloc(m_bitmap.bitmap, (m_blocksCount >> 3) + 1);
                m_bitmap.allocsize = (m_blocksCount >> 3) + 1;
                m_bitmap.bitmap[m_blocksCount >> 3] = 0xff;
        }

        m_bitmap.count = m_blocksCount;
        m_bitmap.offset = 0;

        //  walk through the bitmap and count up every used block
        m_usedBlocks = 0;
        u_int64_t i = 0;
        for (i = 0; i < (m_blocksCount >> 3); ++i ) {
                m_usedBlocks += bits_hash[(u_int8_t)m_bitmap.bitmap[i]];
        }

        u_int64_t lastbits = m_blocksCount & 0x07;
        if ( lastbits != 0 ) {
                //  the last byte in the bitmap
                //  describes only 'lastbits' number of blocks
                //  not 8 blocks
                u_int8_t lastbyte = m_bitmap.bitmap[i];
                m_usedBlocks += bits_hash[(u_int8_t)(lastbyte << (8 - lastbits))];
        }

        cerr << "NtfsPartition info:" << endl;
        cerr << "Block size = " << m_blockSize << endl;
        cerr << "Blocks count = " << m_blocksCount << endl;
        cerr << "Used blocks count = " << m_usedBlocks << endl;
        cerr << "Bitmap data size = " << partimage_ng_ntfs_bitmap_data_size(m_ntfs_vol) << endl;
        cerr << "Bitmap allocated size = " << partimage_ng_ntfs_bitmap_allocated_size(m_ntfs_vol) << endl;

        struct ntfs_badblocks_run *bbr = partimage_ng_ntfs_badclusters(m_ntfs_vol);
        if ( bbr->length ) {
                cerr << "Found blocks marked as bad" << endl;
                struct ntfs_badblocks_run *bbr_cur = bbr;
                for ( ; bbr_cur->length; bbr_cur++ ) {
                        cerr << "Bad block # = " << bbr_cur->start << endl;
                        cerr << "Count = " << bbr_cur->length << endl;
                        for ( u_int64_t i = 0; i < bbr_cur->length; i++ ) {
                                set_bit(&m_bitmap, bbr_cur->start + i, 0);
                                m_usedBlocks--;
                        }
                }
                cerr << "Corrected used blocks count = " << m_usedBlocks << endl;
        }
        free(bbr);
}

NtfsPartition::~NtfsPartition()
{
        if ( partimage_ng_ntfs_close(m_ntfs_vol) != 0 ) {
                cerr << "Couldn't close the filesystem" << endl;
        }

        free(m_bitmap.bitmap);
}

size_t NtfsPartition::blockSize() const
{
        return m_blockSize;
}

u_int64_t NtfsPartition::blocksCount() const
{
        return m_blocksCount;
}

u_int64_t NtfsPartition::usedBlocks() const
{
        return m_usedBlocks;
}

const struct Bitmap* NtfsPartition::bitmap()
{
        return &m_bitmap;
}

bool NtfsPartition::dumpInternalBitmap(int dumpd) const
{
        u_int64_t bitmapSize = partimage_ng_ntfs_bitmap_allocated_size(m_ntfs_vol);
        u_int8_t *bitmap = partimage_ng_ntfs_bitmap(m_ntfs_vol);

        if ( writen(dumpd, bitmap, bitmapSize) == -1 ) {
                perror("Writing the bitmap failed");
                free(bitmap);
                return false;
        }

        free(bitmap);
        return true;
}
