/*
    Copyright (C) 2007    Dmitry Ivanov, vonami@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <unistd.h>
#include <errno.h>
#include <iconv.h>
#include <cstring>
#include <iostream>

#include "partimage-ng.h"

using namespace std;
using namespace PartimageNG;

ssize_t PartimageNG::readn(int fd, const void *buffer, size_t size, IOErrorsHandling errHandling)
{
        u_int8_t *buf = (u_int8_t *) buffer;
        ssize_t readnum = 0;
        size_t toread = size;
        while ( toread != 0 ) {
                readnum = read(fd, buf, toread);
                if ( readnum < 0 ) {
                        if ( errno == EINTR ) {
                                readnum = 0;    // read again
                        }
                        else if ( errno == EIO && errHandling == IgnoreIOErrors ) {      // perhaps a bad sector
                                size_t remainder = 512 - ( (size - toread) % 512 );
                                memset(buf, 0, remainder);
                                if ( lseek(fd, remainder, SEEK_CUR) == -1 ) {
                                        return -1;
                                }
                                toread -= remainder;
                                buf += remainder;
                        }
                        else {
                                return -1;
                        }
                }
                else if ( readnum == 0 ) {
                        break;                  // EOF
                }
                else {
                        toread -= readnum;
                        buf += readnum;
                }
        }

        return (size - toread);
}

ssize_t PartimageNG::writen(int fd, const void *buffer, size_t size, IOErrorsHandling errHandling)
{
        u_int8_t *buf = (u_int8_t *) buffer;
        ssize_t writenum = 0;
        size_t towrite = size;
        while ( towrite != 0 ) {
                writenum = write(fd, buf, towrite);
                if ( writenum <= 0 ) {
                        if ( errno == EINTR ) {
                                writenum = 0;    // write again
                        }
                        else if ( errno == ENOSPC ) {
                                break;          // EOF
                        }
                        else {
                                return -1;
                        }
                }
                else {
                        towrite -= writenum;
                        buf += writenum;
                }
        }

        return (size - towrite);
}

ssize_t PartimageNG::preadn(int fd, const void *buffer, size_t size, off_t offset, IOErrorsHandling errHandling)
{
        if ( lseek(fd, offset, SEEK_SET) == -1 ) {
                return -1;
        }

        return readn(fd, buffer, size, errHandling);
}

ssize_t PartimageNG::pwriten(int fd, const void *buffer, size_t size, off_t offset, IOErrorsHandling errHandling)
{
        if ( lseek(fd, offset, SEEK_SET) == -1 ) {
                return -1;
        }

        return writen(fd, buffer, size, errHandling);
}

bool PartimageNG::test_bit(const struct Bitmap *bitmap, u_int64_t bit)
{
        if ( bit >= bitmap->count )
                return false;

        if ( bitmap->bitmap[(bit + bitmap->offset) >> 3] & (1 << ((bit + bitmap->offset) & 0x07)) )
                return true;
        else
                return false;
}

void PartimageNG::set_bit(struct Bitmap *bitmap, u_int64_t bit, int value)
{
        if ( value )
                bitmap->bitmap[(bit + bitmap->offset) >> 3] |= 1 << (((bit + bitmap->offset) >> 3) & 0x07);
        else
                bitmap->bitmap[(bit + bitmap->offset) >> 3] &= ~(1 << (((bit + bitmap->offset) >> 3) & 0x07));
}

int PartimageNG::convertToUTF8(char **out, size_t *outsize, const char *in, size_t insize, const char *incodeset)
{
        *out = NULL;
        *outsize = 0;

        iconv_t ic = iconv_open("UTF-8", incodeset);
        if ( ic == (iconv_t) -1 ) {
                if ( errno == EINVAL ) {
                        cerr << "Conversion from " << incodeset
                             << "to UTF-8 not available" << endl;
                        return 1;
                }
                else {
                        perror("iconv_open");
                        return 1;
                }
        }

        // temp vars for iconv()
        char *temp_in, *temp_out;
        size_t temp_insize, temp_outsize;
        char *buf = NULL;                       // buffer for conversion
        size_t room = insize;                   // size of this buffer

        // perform conversion until the buffer grows enough
        size_t nconv = 0;
        errno = 0;
        do {
                room = room * 2;                // double the avail room
                temp_insize = insize;
                temp_in = (char *) in;
                temp_outsize = room;
                buf = temp_out = (char *) realloc(buf, room + 1);               // +1 to reserve space for '\0'
                nconv = iconv(ic, &temp_in, &temp_insize, &temp_out, &temp_outsize);
        }
        while ( (nconv == (size_t)-1) && (errno == E2BIG) );

        if ( nconv == (size_t) -1 ) {
                perror("iconv");
                free(buf);
                iconv_close(ic);
                return 1;
        }

        // add the terminating '\0'
        buf[room - temp_outsize] = 0;

        // return the result buffer
        *out = buf;
        *outsize = room - temp_outsize;

        iconv_close(ic);
        return 0;
}

int PartimageNG::convertFromUTF8(char **out, size_t *outsize, const char *in, size_t insize, const char *outcodeset)
{
        *out = NULL;
        *outsize = 0;

        iconv_t ic = iconv_open(outcodeset, "UTF-8");
        if ( ic == (iconv_t) -1 ) {
                if ( errno == EINVAL ) {
                        cerr << "Conversion from UTF-8 to" << outcodeset
                             << "not available" << endl;
                        return 1;
                }
                else {
                        perror("iconv_open");
                        return 1;
                }
        }

        // temp vars for iconv()
        char *temp_in, *temp_out;
        size_t temp_insize, temp_outsize;
        char *buf = NULL;                       // buffer for conversion
        size_t room = insize;                   // size of this buffer

        // perform conversion until the buffer grows enough
        size_t nconv = 0;
        errno = 0;
        do {
                room = room * 2;                // double the avail room
                temp_insize = insize;
                temp_in = (char *) in;
                temp_outsize = room;
                buf = temp_out = (char *) realloc(buf, room + 1);               // +1 to reserve space for '\0'
                nconv = iconv(ic, &temp_in, &temp_insize, &temp_out, &temp_outsize);
        }
        while ( (nconv == (size_t)-1) && (errno == E2BIG) );

        if ( nconv == (size_t) -1 ) {
                perror("iconv");
                free(buf);
                iconv_close(ic);
                return 1;
        }

        // add the terminating '\0'
        buf[room - temp_outsize] = 0;

        // return the result buffer
        *out = buf;
        *outsize = room - temp_outsize;

        iconv_close(ic);
        return 0;
}
